@@ -21,8 +21,8 @@
 
 #include <stdlib.h>
 
-#include <sepol/flask_types.h>
-#include <sepol/conditional.h>
+#include <sepol/policydb/flask_types.h>
+#include <sepol/policydb/conditional.h>
 
 #include "private.h"
 
@@ -34,7 +34,7 @@ static void cond_optimize(cond_av_list_t **l) {
 	top = p = cur = *l;
 
 	while(cur) {
-		if ((cur->node->datum.specified & AVTAB_TYPE) && (top != cur)) {
+		if ((cur->node->key.specified & AVTAB_TYPE) && (top != cur)) {
 			p->next = cur->next;
 			cur->next = top;
 			top = cur;
@@ -58,8 +58,8 @@ void cond_optimize_lists(cond_list_t* cl) {
 }
 
 
-static int bool_present(int target, unsigned int bools[], int num_bools) {
-	int i = 0;
+static int bool_present(unsigned int target, unsigned int bools[], unsigned int num_bools) {
+	unsigned int i = 0;
 	int ret = 1;
 
 	if (num_bools > COND_MAX_BOOLS) {
@@ -71,7 +71,7 @@ static int bool_present(int target, unsigned int bools[], int num_bools) {
 }
 
 static int same_bools(cond_node_t *a, cond_node_t *b) {
-	int i, x;
+	unsigned int i, x;
 	
 	x = a->nbools;
 
@@ -124,37 +124,58 @@ static int cond_expr_equal(cond_node_t *a, cond_node_t *b)
 	return 1;
 }
 
+/* Find a conditional (the needle) within a list of existing ones (the
+ * haystack) that has a matching expression.  If found, return a
+ * pointer to the existing node, setting 'was_created' to 0.
+ * Otherwise create a new one and return it, setting 'was_created' to
+ * 1. */
+cond_node_t *cond_node_find(policydb_t *p,
+                            cond_node_t *needle, cond_node_t *haystack,
+                            int *was_created) {
+	cond_node_t *new_node;
+	unsigned int i;
+
+	while(haystack) {
+		if (cond_expr_equal(needle,haystack)) {
+                        *was_created = 0;
+                        return haystack;
+                }
+		haystack = haystack->next;
+	}
+        *was_created = 1;
+        new_node = (cond_node_t *)malloc(sizeof (cond_node_t));
+        if (!new_node) {
+                return NULL;
+        }
+        memset(new_node, 0, sizeof(cond_node_t));
+        new_node->expr = cond_copy_expr(needle->expr);
+        if (!new_node->expr) {
+		free(new_node);
+                return NULL;
+	}
+        new_node->cur_state = cond_evaluate_expr(p, new_node->expr);
+        new_node->nbools = needle->nbools;
+        for (i = 0; i < needle->nbools; i++)
+                new_node->bool_ids[i] = needle->bool_ids[i];
+        new_node->expr_pre_comp = needle->expr_pre_comp;
+        new_node->true_list = NULL;
+        new_node->false_list = NULL;
+        new_node->avtrue_list = NULL;
+        new_node->avfalse_list = NULL;
+        return new_node;
+}
+
 /* return either a pre-existing matching node or create a new node */
-cond_node_t *cond_node_search(policydb_t *p, cond_node_t *cn)
+cond_node_t *cond_node_search(policydb_t *p, cond_node_t *list, cond_node_t *cn)
 {
-	cond_node_t *list,*new;
-	int i;
-
-	list = p->cond_list;
-
-	while(list) {
-		if (cond_expr_equal(cn,list)) break;
-		list = list->next;
-	}
-	if (list) return list;
-	else {
-		new = (cond_node_t *)malloc(sizeof (cond_node_t));
-		if (!new) {
-			return NULL;
-		}
-		new->expr = cn->expr;
-		new->cur_state = cond_evaluate_expr(p, new->expr);
-		new->nbools = cn->nbools;
-		for (i = 0; i < cn->nbools; i++)
-			new->bool_ids[i] = cn->bool_ids[i];
-		new->expr_pre_comp = cn->expr_pre_comp;
-		new->true_list = NULL;
-		new->false_list = NULL;
-		/* add conditional node to policy list */
-		new->next = p->cond_list;
-		p->cond_list = new;
-		return new;
-	}
+        int was_created;
+        cond_node_t *result = cond_node_find(p, cn, list, &was_created);
+        if (result != NULL && was_created) {
+                /* add conditional node to policy list */
+                result->next = p->cond_list;
+                p->cond_list = result;
+        }
+        return result;
 }
 
 /*
@@ -220,6 +241,38 @@ int cond_evaluate_expr(policydb_t *p, cond_expr_t *expr)
 	return s[0];
 }
 
+cond_expr_t *cond_copy_expr(cond_expr_t *expr)
+{
+        cond_expr_t *cur, *head, *tail, *new_expr;
+        tail = head = NULL;
+        cur = expr;
+        while (cur) {
+                new_expr = (cond_expr_t*)malloc(sizeof(cond_expr_t));
+                if (!new_expr)
+			goto free_head;
+                memset(new_expr, 0, sizeof(cond_expr_t));
+
+                new_expr->expr_type = cur->expr_type;
+                new_expr->bool = cur->bool;
+
+                if (!head)
+                        head = new_expr;
+                if (tail)
+                        tail->next = new_expr;
+                tail = new_expr;
+                cur = cur->next;
+        }
+        return head;
+
+free_head:
+	while (head) {
+		tail = head->next;
+		free(head);
+		head = tail;
+	}
+	return NULL;
+}
+
 /*
  * evaluate_cond_node evaluates the conditional stored in
  * a cond_node_t and if the result is different than the
@@ -240,18 +293,18 @@ static int evaluate_cond_node(policydb_t *p, cond_node_t *node)
 		/* turn the rules on or off */
 		for (cur = node->true_list; cur != NULL; cur = cur->next) {
 			if (new_state <= 0) {
-				cur->node->datum.specified &= ~AVTAB_ENABLED;
+				cur->node->key.specified &= ~AVTAB_ENABLED;
 			} else {
-				cur->node->datum.specified |= AVTAB_ENABLED;
+				cur->node->key.specified |= AVTAB_ENABLED;
 			}
 		}
 
 		for (cur = node->false_list; cur != NULL; cur = cur->next) {	
 			/* -1 or 1 */
 			if (new_state) {
-				cur->node->datum.specified &= ~AVTAB_ENABLED;
+				cur->node->key.specified &= ~AVTAB_ENABLED;
 			} else {
-				cur->node->datum.specified |= AVTAB_ENABLED;
+				cur->node->key.specified |= AVTAB_ENABLED;
 			}
 		}
 	}
@@ -266,9 +319,10 @@ int cond_normalize_expr(policydb_t *p, cond_node_t *cn)
 {
 	cond_expr_t *ne,*e;
 	cond_av_list_t *tmp;
-	int i, j, k, orig_value[COND_MAX_BOOLS];
+	unsigned int i, j, orig_value[COND_MAX_BOOLS];
+	int k;
 	uint32_t test = 0x0;
-	
+	avrule_t *tmp2;
 
 	cn->nbools = 0;
 
@@ -294,6 +348,10 @@ int cond_normalize_expr(policydb_t *p, cond_node_t *cn)
 		tmp = cn->true_list;
 		cn->true_list = cn->false_list;
 		cn->false_list = tmp;
+		tmp2 = cn->avtrue_list;
+		cn->avtrue_list = cn->avfalse_list;
+		cn->avfalse_list = tmp2;
+                
 		/* free the "not" node in the list */
 		free(e);
 	}
@@ -325,7 +383,7 @@ int cond_normalize_expr(policydb_t *p, cond_node_t *cn)
 		}
 		
 		/* loop through all possible combinations of values for bools in expression */
-		for(test = 0x0; test < (0x1 << cn->nbools) ; test++) {
+		for(test = 0x0; test < (0x1U << cn->nbools) ; test++) {
 			/* temporarily set the value for all the bools in the
 			 * expression using the corr.  bit in test */
 			for(j = 0; j < cn->nbools; j++) {
@@ -372,7 +430,7 @@ int cond_policydb_init(policydb_t *p)
 	return 0;
 }
 
-static void cond_av_list_destroy(cond_av_list_t *list)
+void cond_av_list_destroy(cond_av_list_t *list)
 {
 	cond_av_list_t *cur, *next;
 	for (cur = list; cur != NULL; cur = next) {
@@ -382,23 +440,31 @@ static void cond_av_list_destroy(cond_av_list_t *list)
 	}
 }
 
-static void cond_node_destroy(cond_node_t *node)
-{
+void cond_expr_destroy(cond_expr_t *expr) {
 	cond_expr_t *cur_expr, *next_expr;
 
-	if (!node)
+	if (!expr)
 		return;
 
-	for (cur_expr = node->expr; cur_expr != NULL; cur_expr = next_expr) {
+	for (cur_expr = expr; cur_expr != NULL; cur_expr = next_expr) {
 		next_expr = cur_expr->next;
 		free(cur_expr);
 	}
-	cond_av_list_destroy(node->true_list);
-	cond_av_list_destroy(node->false_list);
-	free(node);
 }
 
-static void cond_list_destroy(cond_list_t *list)
+void cond_node_destroy(cond_node_t *node)
+{
+	if (!node)
+		return;
+
+	cond_expr_destroy(node->expr);
+        avrule_list_destroy(node->avtrue_list);
+        avrule_list_destroy(node->avfalse_list);
+	cond_av_list_destroy(node->true_list);
+	cond_av_list_destroy(node->false_list);
+}
+
+void cond_list_destroy(cond_list_t *list)
 {
 	cond_node_t *next, *cur;
 	
@@ -408,6 +474,7 @@ static void cond_list_destroy(cond_list_t *list)
 	for (cur = list; cur != NULL; cur = next) {
 		next = cur->next;
 		cond_node_destroy(cur);
+                free(cur);
 	}
 }
 
@@ -430,7 +497,7 @@ int cond_init_bool_indexes(policydb_t *p)
 	return 0;
 }
 
-int cond_destroy_bool(hashtab_key_t key, hashtab_datum_t datum, void *p)
+int cond_destroy_bool(hashtab_key_t key, hashtab_datum_t datum, void *p __attribute__ ((unused)))
 {
 	if (key)
 		free(key);
@@ -462,7 +529,7 @@ static int bool_isvalid(cond_bool_datum_t *b)
 	return 1;
 }
 
-int cond_read_bool(policydb_t *p, hashtab_t h, struct policy_file *fp)
+int cond_read_bool(policydb_t *p __attribute__ ((unused)), hashtab_t h, struct policy_file *fp)
 {
 	char *key = 0;
 	cond_bool_datum_t *booldatum;
@@ -502,15 +569,100 @@ err:
 	return -1;
 }
 
+struct cond_insertf_data 
+{
+	struct policydb *p;
+	cond_av_list_t *other;
+	cond_av_list_t *head;
+	cond_av_list_t *tail;
+};
+
+static int cond_insertf(avtab_t *a __attribute__ ((unused)), avtab_key_t *k, avtab_datum_t *d, void *ptr)
+{
+	struct cond_insertf_data *data = ptr;
+	struct policydb *p = data->p;
+	cond_av_list_t *other = data->other, *list, *cur;
+	avtab_ptr_t node_ptr;
+	uint8_t found;
+
+
+	/*
+	 * For type rules we have to make certain there aren't any
+	 * conflicting rules by searching the te_avtab and the
+	 * cond_te_avtab.
+	 */
+	if (k->specified & AVTAB_TYPE) {
+		if (avtab_search(&p->te_avtab, k)) {
+			printf("security: type rule already exists outside of a conditional.");
+			goto err;
+		}
+		/*
+		 * If we are reading the false list other will be a pointer to
+		 * the true list. We can have duplicate entries if there is only
+		 * 1 other entry and it is in our true list.
+		 *
+		 * If we are reading the true list (other == NULL) there shouldn't
+		 * be any other entries.
+		 */
+		if (other) {
+			node_ptr = avtab_search_node(&p->te_cond_avtab, k);
+			if (node_ptr) {
+				if (avtab_search_node_next(node_ptr, k->specified)) {
+					printf("security: too many conflicting type rules.");
+					goto err;
+				}
+				found = 0;
+				for (cur = other; cur != NULL; cur = cur->next) {
+					if (cur->node == node_ptr) {
+						found = 1;
+						break;
+					}
+				}
+				if (!found) {
+					printf("security: conflicting type rules.\n");
+					goto err;
+				}
+			}
+		} else {    
+			if (avtab_search(&p->te_cond_avtab, k)) {
+				printf("security: conflicting type rules when adding type rule for true.\n");
+				goto err;
+			}
+		}
+	}
+
+	node_ptr = avtab_insert_nonunique(&p->te_cond_avtab, k, d);
+	if (!node_ptr) {
+		printf("security: could not insert rule.");
+		goto err;
+	}
+	node_ptr->parse_context = (void*)1;
+
+	list = malloc(sizeof(cond_av_list_t));
+	if (!list)
+		goto err;
+	memset(list, 0, sizeof(cond_av_list_t));
+
+	list->node = node_ptr;
+	if (!data->head)
+		data->head = list;
+	else
+		data->tail->next = list;
+	data->tail = list;
+	return 0;
+
+err:
+	cond_av_list_destroy(data->head);
+	data->head = NULL;
+	return -1;
+}
+
 static int cond_read_av_list(policydb_t *p, void *fp, cond_av_list_t **ret_list, cond_av_list_t *other)
 {
-	cond_av_list_t *list, *last = NULL, *cur;
-	avtab_key_t key;
-	avtab_datum_t datum;
-	avtab_ptr_t node_ptr;
-	int len, i;
-	uint32_t *buf;
-	uint8_t found;
+	unsigned int i;
+	int rc;
+	uint32_t *buf, len;
+	struct cond_insertf_data data;
 
 	*ret_list = NULL;
 
@@ -524,80 +676,19 @@ static int cond_read_av_list(policydb_t *p, void *fp, cond_av_list_t **ret_list,
 		return 0;
 	}
 
+	data.p = p;
+	data.other = other;
+	data.head = NULL;
+	data.tail = NULL;
 	for (i = 0; i < len; i++) {
-		if (avtab_read_item(fp, &datum, &key))
-			goto err;
-
-		/*
-		 * For type rules we have to make certain there aren't any
-		 * conflicting rules by searching the te_avtab and the
-		 * cond_te_avtab.
-		 */
-		if (datum.specified & AVTAB_TYPE) {
-			if (avtab_search(&p->te_avtab, &key, AVTAB_TYPE)) {
-				printf("security: type rule already exists outside of a conditional.");
-				goto err;
-			}
-			/*
-			 * If we are reading the false list other will be a pointer to
-			 * the true list. We can have duplicate entries if there is only
-			 * 1 other entry and it is in our true list.
-			 *
-			 * If we are reading the true list (other == NULL) there shouldn't
-			 * be any other entries.
-			 */
-			if (other) {
-				node_ptr = avtab_search_node(&p->te_cond_avtab, &key, AVTAB_TYPE);
-				if (node_ptr) {
-					if (avtab_search_node_next(node_ptr, AVTAB_TYPE)) {
-						printf("security: too many conflicting type rules.");
-						goto err;
-					}
-					found = 0;
-					for (cur = other; cur != NULL; cur = cur->next) {
-						if (cur->node == node_ptr) {
-							found = 1;
-							break;
-						}
-					}
-					if (!found) {
-						printf("security: conflicting type rules.\n");
-						goto err;
-					}
-				}
-			} else {    
-				if (avtab_search(&p->te_cond_avtab, &key, AVTAB_TYPE)) {
-					printf("security: conflicting type rules when adding type rule for true.\n");
-					goto err;
-				}
-			}
-		}
-		node_ptr = avtab_insert_nonunique(&p->te_cond_avtab, &key, &datum);
-		if (!node_ptr) {
-			printf("security: could not insert rule.");
-			goto err;
-		}
-		node_ptr->parse_context = (void*)1;
-
-		list = malloc(sizeof(cond_av_list_t));
-		if (!list)
-			goto err;
-		memset(list, 0, sizeof(cond_av_list_t));
-
-		list->node = node_ptr;
-		if (i == 0)
-			*ret_list = list;
-		else
-			last->next = list;
-		last = list;
+		rc = avtab_read_item(fp, p->policyvers, &p->te_cond_avtab, cond_insertf, &data);
+		if (rc)
+			return rc;
 
 	}
 
+	*ret_list = data.head;
 	return 0;
-err:
-	cond_av_list_destroy(*ret_list);
-	*ret_list = NULL;
-	return -1;
 }
 
 static int expr_isvalid(policydb_t *p, cond_expr_t *expr)
@@ -622,14 +713,14 @@ static int cond_read_node(policydb_t *p, cond_node_t *node, void *fp)
 
 	buf = next_entry(fp, sizeof(uint32_t));
 	if (!buf)
-		return -1;
+		goto err;
 
 	node->cur_state = le32_to_cpu(buf[0]);
 	
 	len = 0;
 	buf = next_entry(fp, sizeof(uint32_t));
 	if (!buf)
-		return -1;
+		goto err;
 
 	/* expr */
 	len = le32_to_cpu(buf[0]);
@@ -648,8 +739,10 @@ static int cond_read_node(policydb_t *p, cond_node_t *node, void *fp)
 		expr->expr_type = le32_to_cpu(buf[0]);
 		expr->bool = le32_to_cpu(buf[1]);
 		
-		if (!expr_isvalid(p, expr))
+		if (!expr_isvalid(p, expr)) {
+			free(expr);
 			goto err;
+		}
 
 		if (i == 0) {
 			node->expr = expr;
@@ -659,17 +752,26 @@ static int cond_read_node(policydb_t *p, cond_node_t *node, void *fp)
 		last = expr;
 	}
 	
-	if (cond_read_av_list(p, fp, &node->true_list, NULL) != 0)
-		goto err;
-	if (cond_read_av_list(p, fp, &node->false_list, node->true_list) != 0)
-		goto err;
+	if (p->policy_type == POLICY_KERN) {
+		if (cond_read_av_list(p, fp, &node->true_list, NULL) != 0)
+			goto err;
+		if (cond_read_av_list(p, fp, &node->false_list, node->true_list) != 0)
+			goto err;
+	} else {
+		if (avrule_read_list(p, &node->avtrue_list, fp))
+			goto err;
+		if (avrule_read_list(p, &node->avfalse_list, fp))
+			goto err;
+	}
+        
 	return 0;
 err:
 	cond_node_destroy(node);
+	free(node);
 	return -1;
 }
 
-int cond_read_list(policydb_t *p, void *fp)
+int cond_read_list(policydb_t *p, cond_list_t **list, void *fp)
 {
 	cond_node_t *node, *last = NULL;
 	uint32_t *buf;
@@ -691,7 +793,7 @@ int cond_read_list(policydb_t *p, void *fp)
 			goto err;
 
 		if (i == 0) {
-			p->cond_list = node;
+			*list = node;
 		} else {
 			last->next = node;
 		}
@@ -712,22 +814,22 @@ void cond_compute_av(avtab_t *ctab, avtab_key_t *key, struct sepol_av_decision *
 	if(!ctab || !key || !avd) 
 		return;
 		
-	for(node = avtab_search_node(ctab, key, AVTAB_AV); node != NULL; 
-				node = avtab_search_node_next(node, AVTAB_AV)) {
-		if ( (uint32_t) (AVTAB_ALLOWED|AVTAB_ENABLED) ==
-		     (node->datum.specified & (AVTAB_ALLOWED|AVTAB_ENABLED)))
-			avd->allowed |= avtab_allowed(&node->datum);
-		if ( (uint32_t) (AVTAB_AUDITDENY|AVTAB_ENABLED) ==
-		     (node->datum.specified & (AVTAB_AUDITDENY|AVTAB_ENABLED)))
+	for(node = avtab_search_node(ctab, key); node != NULL; 
+				node = avtab_search_node_next(node, key->specified)) {
+		if ( (uint16_t) (AVTAB_ALLOWED|AVTAB_ENABLED) ==
+		     (node->key.specified & (AVTAB_ALLOWED|AVTAB_ENABLED)))
+			avd->allowed |= node->datum.data;
+		if ( (uint16_t) (AVTAB_AUDITDENY|AVTAB_ENABLED) ==
+		     (node->key.specified & (AVTAB_AUDITDENY|AVTAB_ENABLED)))
 			/* Since a '0' in an auditdeny mask represents a 
 			 * permission we do NOT want to audit (dontaudit), we use
 			 * the '&' operand to ensure that all '0's in the mask
 			 * are retained (much unlike the allow and auditallow cases).
 			 */
-			avd->auditdeny &= avtab_auditdeny(&node->datum);
-		if ( (uint32_t) (AVTAB_AUDITALLOW|AVTAB_ENABLED) ==
-		     (node->datum.specified & (AVTAB_AUDITALLOW|AVTAB_ENABLED)))
-			avd->auditallow |= avtab_auditallow(&node->datum);
+			avd->auditdeny &= node->datum.data;
+		if ( (uint16_t) (AVTAB_AUDITALLOW|AVTAB_ENABLED) ==
+		     (node->key.specified & (AVTAB_AUDITALLOW|AVTAB_ENABLED)))
+			avd->auditallow |= node->datum.data;
 	}
 	return;	
 }
