@@ -49,32 +49,34 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
-#include <sepol/context.h>
-#include <sepol/policydb.h>
-#include <sepol/sidtab.h>
-#include <sepol/services.h>
-#include <sepol/mls.h>
-#include <sepol/conditional.h>
-#include <sepol/flask.h>
+#include <sepol/policydb/policydb.h>
+#include <sepol/policydb/sidtab.h>
+#include <sepol/policydb/services.h>
+#include <sepol/policydb/conditional.h>
+#include <sepol/policydb/flask.h>
 
+#include "debug.h"
 #include "private.h"
+#include "context.h"
 #include "av_permissions.h"
+#include "dso.h"
+#include "mls.h"
 
-#define BUG() do { printf("Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); } while (0)
-#define BUG_ON(x) do { if (x) printf("Badness in %s at %s:%d\n", __FUNCTION__, __FILE__, __LINE__); } while (0)
+#define BUG() do { ERR(NULL, "Badness at %s:%d", __FILE__, __LINE__); } while (0)
+#define BUG_ON(x) do { if (x) ERR(NULL, "Badness at %s:%d", __FILE__, __LINE__); } while (0)
 
 static int selinux_enforcing = 1;
 
 static sidtab_t mysidtab, *sidtab = &mysidtab;
 static policydb_t mypolicydb, *policydb = &mypolicydb;
 
-int sepol_set_sidtab(sidtab_t *s) 
+int hidden sepol_set_sidtab(sidtab_t *s) 
 {
 	sidtab = s;
 	return 0;
 }
 
-int sepol_set_policydb(policydb_t *p) 
+int hidden sepol_set_policydb(policydb_t *p) 
 {
 	policydb = p;
 	return 0;
@@ -85,8 +87,14 @@ int sepol_set_policydb_from_file(FILE *fp)
 	struct policy_file pf;
 	pf.fp = fp;
 	pf.type = PF_USE_STDIO;
+	if (mypolicydb.policy_type)
+		policydb_destroy(&mypolicydb);
+	if (policydb_init(&mypolicydb)) {
+		ERR(NULL, "Out of memory!");
+		return -1;
+	}
 	if (policydb_read(&mypolicydb, &pf, 0)) {
-		fprintf(stderr, "Can't read binary policy:  %s\n",
+		ERR(NULL, "can't read binary policy: %s",
 			strerror(errno));
 		return -1;
 	}
@@ -114,7 +122,7 @@ static uint32_t latest_granting = 0;
  * of the process performing the transition.  All other callers of
  * constraint_expr_eval should pass in NULL for xcontext.
  */
-int constraint_expr_eval(context_struct_t *scontext,
+static int constraint_expr_eval(context_struct_t *scontext,
                          context_struct_t *tcontext,
                          context_struct_t *xcontext,
                          constraint_expr_t *cexpr)
@@ -303,12 +311,14 @@ static int context_struct_compute_av(context_struct_t *scontext,
 	constraint_node_t *constraint;
 	struct role_allow *ra;
 	avtab_key_t avkey;
-	avtab_datum_t *avdatum;
 	class_datum_t *tclass_datum;
+	avtab_ptr_t node;
+	ebitmap_t *sattr, *tattr;
+	ebitmap_node_t *snode, *tnode;
+	unsigned int i, j;
 
 	if (!tclass || tclass > policydb->p_classes.nprim) {
-		printf("security_compute_av:  unrecognized class %d\n",
-		       tclass);
+		ERR(NULL, "unrecognized class %d", tclass);
 		return -EINVAL;
 	}
 	tclass_datum = policydb->class_val_to_struct[tclass - 1];
@@ -327,21 +337,34 @@ static int context_struct_compute_av(context_struct_t *scontext,
 	 * If a specific type enforcement rule was defined for
 	 * this permission check, then use it.
 	 */
-	avkey.source_type = scontext->type;
-	avkey.target_type = tcontext->type;
 	avkey.target_class = tclass;
-	avdatum = avtab_search(&policydb->te_avtab, &avkey, AVTAB_AV);
-	if (avdatum) {
-		if (avdatum->specified & AVTAB_ALLOWED)
-			avd->allowed = avtab_allowed(avdatum);
-		if (avdatum->specified & AVTAB_AUDITDENY)
-			avd->auditdeny = avtab_auditdeny(avdatum);
-		if (avdatum->specified & AVTAB_AUDITALLOW)
-			avd->auditallow = avtab_auditallow(avdatum);
-	}
+	avkey.specified = AVTAB_AV;
+	sattr = &policydb->type_attr_map[scontext->type - 1];
+	tattr = &policydb->type_attr_map[tcontext->type - 1];
+	ebitmap_for_each_bit(sattr, snode, i) {
+		if (!ebitmap_node_get_bit(snode, i))
+			continue;
+		ebitmap_for_each_bit(tattr, tnode, j) {
+			if (!ebitmap_node_get_bit(tnode, j))
+				continue;
+			avkey.source_type = i + 1;
+			avkey.target_type = j + 1;
+			for (node = avtab_search_node(&policydb->te_avtab, &avkey);
+			     node != NULL;
+			     node = avtab_search_node_next(node, avkey.specified)) {
+				if (node->key.specified == AVTAB_ALLOWED)
+					avd->allowed |= node->datum.data;
+				else if (node->key.specified == AVTAB_AUDITALLOW)
+					avd->auditallow |= node->datum.data;
+				else if (node->key.specified == AVTAB_AUDITDENY)
+					avd->auditdeny &= node->datum.data;
+			}
 
-	/* Check conditional av table for additional permissions */
-	cond_compute_av(&policydb->te_cond_avtab, &avkey, avd);
+			/* Check conditional av table for additional permissions */
+			cond_compute_av(&policydb->te_cond_avtab, &avkey, avd);
+
+		}
+	}
 
 	if (requested & ~avd->allowed) {
 		*reason |= SEPOL_COMPUTEAV_TE;
@@ -393,7 +416,7 @@ static int context_struct_compute_av(context_struct_t *scontext,
 	return 0;
 }
 
-int sepol_validate_transition(sepol_security_id_t oldsid, sepol_security_id_t newsid,
+int hidden sepol_validate_transition(sepol_security_id_t oldsid, sepol_security_id_t newsid,
                               sepol_security_id_t tasksid, sepol_security_class_t tclass)
 {
 	context_struct_t *ocontext;
@@ -403,30 +426,26 @@ int sepol_validate_transition(sepol_security_id_t oldsid, sepol_security_id_t ne
 	constraint_node_t *constraint;
 
 	if (!tclass || tclass > policydb->p_classes.nprim) {
-		printf("sepol_validate_transition:  "
-		       "unrecognized class %d\n", tclass);
+		ERR(NULL, "unrecognized class %d", tclass);
 		return -EINVAL;
 	}
 	tclass_datum = policydb->class_val_to_struct[tclass - 1];
 
 	ocontext = sepol_sidtab_search(sidtab, oldsid);
 	if (!ocontext) {
-		printf("sepol_validate_transition: "
-		       " unrecognized SID %d\n", oldsid);
+		ERR(NULL, "unrecognized SID %d", oldsid);
 		return -EINVAL;
 	}
 
 	ncontext = sepol_sidtab_search(sidtab, newsid);
 	if (!ncontext) {
-		printf("sepol_validate_transition: "
-		       " unrecognized SID %d\n", newsid);
+		ERR(NULL, "unrecognized SID %d", newsid);
 		return -EINVAL;
 	}
 
 	tcontext = sepol_sidtab_search(sidtab, tasksid);
 	if (!tcontext) {
-		printf("sepol_validate_transition: "
-		       " unrecognized SID %d\n", tasksid);
+		ERR(NULL, "unrecognized SID %d", tasksid);
 		return -EINVAL;
 	}
 
@@ -442,7 +461,7 @@ int sepol_validate_transition(sepol_security_id_t oldsid, sepol_security_id_t ne
 	return 0;
 }
 
-int sepol_compute_av_reason(sepol_security_id_t ssid,
+int hidden sepol_compute_av_reason(sepol_security_id_t ssid,
 			    sepol_security_id_t tsid,
 			    sepol_security_class_t tclass,
 			    sepol_access_vector_t requested,
@@ -454,13 +473,13 @@ int sepol_compute_av_reason(sepol_security_id_t ssid,
 
 	scontext = sepol_sidtab_search(sidtab, ssid);
 	if (!scontext) {
-		printf("security_compute_av:  unrecognized SID %d\n", ssid);
+		ERR(NULL, "unrecognized SID %d", ssid);
 		rc = -EINVAL;
 		goto out;
 	}
 	tcontext = sepol_sidtab_search(sidtab, tsid);
 	if (!tcontext) {
-		printf("security_compute_av:  unrecognized SID %d\n", tsid);
+		ERR(NULL, "unrecognized SID %d", tsid);
 		rc = -EINVAL;
 		goto out;
 	}
@@ -471,7 +490,7 @@ out:
 	return rc;
 }
 
-int sepol_compute_av(sepol_security_id_t ssid,
+int hidden sepol_compute_av(sepol_security_id_t ssid,
 		     sepol_security_id_t tsid,
 		     sepol_security_class_t tclass,
 		     sepol_access_vector_t requested,
@@ -481,52 +500,6 @@ int sepol_compute_av(sepol_security_id_t ssid,
 	return sepol_compute_av_reason(ssid, tsid, tclass, requested, avd, &reason);
 }
 
-/*
- * Write the security context string representation of 
- * the context structure `context' into a dynamically
- * allocated string of the correct size.  Set `*scontext'
- * to point to this string and set `*scontext_len' to
- * the length of the string.
- */
-int context_struct_to_string(context_struct_t * context,
-			     sepol_security_context_t * scontext,
-			     size_t *scontext_len)
-{
-	char *scontextp;
-
-	*scontext = 0;
-	*scontext_len = 0;
-
-	/* Compute the size of the context. */
-	*scontext_len += strlen(policydb->p_user_val_to_name[context->user - 1]) + 1;
-	*scontext_len += strlen(policydb->p_role_val_to_name[context->role - 1]) + 1;
-	*scontext_len += strlen(policydb->p_type_val_to_name[context->type - 1]) + 1;
-	*scontext_len += mls_compute_context_len(policydb, context);
-
-	/* Allocate space for the context; caller must free this space. */
-	scontextp = malloc(*scontext_len);
-	if (!scontextp) {
-		return -ENOMEM;
-	}
-	*scontext = (sepol_security_context_t) scontextp;
-
-	/*
-	 * Copy the user name, role name and type name into the context.
-	 */
-	sprintf(scontextp, "%s:%s:%s",
-		policydb->p_user_val_to_name[context->user - 1],
-		policydb->p_role_val_to_name[context->role - 1],
-		policydb->p_type_val_to_name[context->type - 1]);
-	scontextp += strlen(policydb->p_user_val_to_name[context->user - 1]) + 1 + strlen(policydb->p_role_val_to_name[context->role - 1]) + 1 + strlen(policydb->p_type_val_to_name[context->type - 1]);
-
-	mls_sid_to_context(policydb, context, &scontextp);
-
-	*scontextp = 0;
-
-	return 0;
-}
-
-
 /*
  * Write the security context string representation of 
  * the context associated with `sid' into a dynamically
@@ -534,7 +507,7 @@ int context_struct_to_string(context_struct_t * context,
  * to point to this string and set `*scontext_len' to
  * the length of the string.
  */
-int sepol_sid_to_context(sepol_security_id_t sid,
+int hidden sepol_sid_to_context(sepol_security_id_t sid,
 			    sepol_security_context_t * scontext,
 			    size_t *scontext_len)
 {
@@ -543,11 +516,11 @@ int sepol_sid_to_context(sepol_security_id_t sid,
 
 	context = sepol_sidtab_search(sidtab, sid);
 	if (!context) {
-		printf("security_sid_to_context:  unrecognized SID %d\n", sid);
+		ERR(NULL, "unrecognized SID %d", sid);
 		rc = -EINVAL;
 		goto out;
 	}
-	rc = context_struct_to_string(context, scontext, scontext_len);
+	rc = context_to_string(NULL, policydb, context, scontext, scontext_len);
 out:
 	return rc;
 	
@@ -557,109 +530,32 @@ out:
  * Return a SID associated with the security context that
  * has the string representation specified by `scontext'.
  */
-int sepol_context_to_sid(sepol_security_context_t scontext,
+int hidden sepol_context_to_sid(sepol_security_context_t scontext,
 			    size_t scontext_len,
 			    sepol_security_id_t * sid)
 {
-	sepol_security_context_t scontext2;
-	context_struct_t context;
-	role_datum_t *role;
-	type_datum_t *typdatum;
-	user_datum_t *usrdatum;
-	char *scontextp, *p, oldc;
-	int rc = 0;
+	context_struct_t* context = NULL;
 
-	if (sid)
-		*sid = SEPOL_SECSID_NULL;
+	/* First, create the context */
+	if (context_from_string(NULL, policydb, &context, 
+		scontext, scontext_len) < 0)
+		goto err;
 
-	/* Copy the string so that we can modify the copy as we parse it.
-	   The string should already by null terminated, but we append a
-	   null suffix to the copy to avoid problems with the existing
-	   attr package, which doesn't view the null terminator as part
-	   of the attribute value. */
-	scontext2 = malloc(scontext_len+1);
-	if (!scontext2) {
-		return -ENOMEM;
-	}
-	memcpy(scontext2, scontext, scontext_len);
-	scontext2[scontext_len] = 0;
+	/* Obtain the new sid */
+	if (sid && (sepol_sidtab_context_to_sid(sidtab, context, sid) < 0))
+		goto err;		
 
-	context_init(&context);
+	context_destroy(context);
+        free(context);
+	return STATUS_SUCCESS;
 
-	/* Parse the security context. */
-
-	rc = -EINVAL;
-	scontextp = (char *) scontext2;
-
-	/* Extract the user. */
-	p = scontextp;
-	while (*p && *p != ':')
-		p++;
-
-	if (*p == 0)
-		goto out;
-
-	*p++ = 0;
-
-	usrdatum = (user_datum_t *) hashtab_search(policydb->p_users.table,
-					      (hashtab_key_t) scontextp);
-	if (!usrdatum)
-		goto out;
-
-	context.user = usrdatum->value;
-
-	/* Extract role. */
-	scontextp = p;
-	while (*p && *p != ':')
-		p++;
-
-	if (*p == 0)
-		goto out;
-
-	*p++ = 0;
-
-	role = (role_datum_t *) hashtab_search(policydb->p_roles.table,
-					       (hashtab_key_t) scontextp);
-	if (!role)
-		goto out;
-	context.role = role->value;
-
-	/* Extract type. */
-	scontextp = p;
-	while (*p && *p != ':')
-		p++;
-	oldc = *p;
-	*p++ = 0;
-
-	typdatum = (type_datum_t *) hashtab_search(policydb->p_types.table,
-					      (hashtab_key_t) scontextp);
-
-	if (!typdatum)
-		goto out;
-
-	context.type = typdatum->value;
-
-	rc = mls_context_to_sid(policydb, oldc, &p, &context);
-	if (rc)
-		goto out;
-
-	if ((p - scontext2) < scontext_len) {
-		rc = -EINVAL;
-		goto out;
-	}
-
-	/* Check the validity of the new context. */
-	if (!policydb_context_isvalid(policydb, &context)) {
-		rc = -EINVAL;
-		goto out;
-	}
-	/* Obtain the new sid. */
-	if (sid)
-		rc = sepol_sidtab_context_to_sid(sidtab, &context, sid);
-out:
-	context_destroy(&context);
-	free(scontext2);
-	return rc;
+	err:
+	if (context) {
+		context_destroy(context);
+                free(context);
+        }
+	ERR(NULL, "could not convert %s to sid", scontext);
+	return STATUS_ERR;
 }
 
 int sepol_check_context(char *context) 
@@ -679,13 +575,12 @@ static inline int compute_sid_handle_invalid_context(
 		sepol_security_context_t s, t, n;
 		size_t slen, tlen, nlen;
 
-		context_struct_to_string(scontext, &s, &slen);
-		context_struct_to_string(tcontext, &t, &tlen);
-		context_struct_to_string(newcontext, &n, &nlen);
-		printf("security_compute_sid:  invalid context %s", n);
-		printf(" for scontext=%s", s);
-		printf(" tcontext=%s", t);
-		printf(" tclass=%s\n", policydb->p_class_val_to_name[tclass-1]);
+		context_to_string(NULL, policydb, scontext, &s, &slen);
+		context_to_string(NULL, policydb, tcontext, &t, &tlen);
+		context_to_string(NULL, policydb, newcontext, &n, &nlen);
+		ERR(NULL, "invalid context %s for "
+			"scontext=%s tcontext=%s tclass=%s",
+			n, s, t, policydb->p_class_val_to_name[tclass-1]);
 		free(s);
 		free(t);
 		free(n);
@@ -704,18 +599,17 @@ static int sepol_compute_sid(sepol_security_id_t ssid,
 	avtab_key_t avkey;
 	avtab_datum_t *avdatum;
 	avtab_ptr_t node;
-	unsigned int type_change = 0;
 	int rc = 0;
 
 	scontext = sepol_sidtab_search(sidtab, ssid);
 	if (!scontext) {
-		printf("security_compute_sid:  unrecognized SID %d\n", ssid);
+		ERR(NULL, "unrecognized SID %d", ssid);
 		rc = -EINVAL;
 		goto out;
 	}
 	tcontext = sepol_sidtab_search(sidtab, tsid);
 	if (!tcontext) {
-		printf("security_compute_sid:  unrecognized SID %d\n", tsid);
+		ERR(NULL, "unrecognized SID %d", tsid);
 		rc = -EINVAL;
 		goto out;
 	}
@@ -753,33 +647,23 @@ static int sepol_compute_sid(sepol_security_id_t ssid,
 	avkey.source_type = scontext->type;
 	avkey.target_type = tcontext->type;
 	avkey.target_class = tclass;
-	avdatum = avtab_search(&policydb->te_avtab, &avkey, AVTAB_TYPE);
+	avkey.specified = specified;
+	avdatum = avtab_search(&policydb->te_avtab, &avkey);
 
 	/* If no permanent rule, also check for enabled conditional rules */
 	if(!avdatum) {
-		node = avtab_search_node(&policydb->te_cond_avtab, &avkey, specified);
+		node = avtab_search_node(&policydb->te_cond_avtab, &avkey);
 		for (; node != NULL; node = avtab_search_node_next(node, specified)) {
-			if (node->datum.specified & AVTAB_ENABLED) {
+			if (node->key.specified & AVTAB_ENABLED) {
 				avdatum = &node->datum;
 				break;
 			}
 		}
 	}
 
-	type_change = (avdatum && (avdatum->specified & specified));
-	if (type_change) {
+	if (avdatum) {
 		/* Use the type from the type transition/member/change rule. */
-		switch (specified) {
-		case AVTAB_TRANSITION:
-			newcontext.type = avtab_transition(avdatum);
-			break;
-		case AVTAB_MEMBER:
-			newcontext.type = avtab_member(avdatum);
-			break;
-		case AVTAB_CHANGE:
-			newcontext.type = avtab_change(avdatum);
-			break;
-		}
+		newcontext.type = avdatum->data;
 	}
 
 	/* Check for class-specific changes. */
@@ -828,7 +712,7 @@ out:
  * Compute a SID to use for labeling a new object in the 
  * class `tclass' based on a SID pair.  
  */
-int sepol_transition_sid(sepol_security_id_t ssid,
+int hidden sepol_transition_sid(sepol_security_id_t ssid,
 			    sepol_security_id_t tsid,
 			    sepol_security_class_t tclass,
 			    sepol_security_id_t * out_sid)
@@ -842,7 +726,7 @@ int sepol_transition_sid(sepol_security_id_t ssid,
  * polyinstantiated object of class `tclass' based on 
  * a SID pair.
  */
-int sepol_member_sid(sepol_security_id_t ssid,
+int hidden sepol_member_sid(sepol_security_id_t ssid,
 			sepol_security_id_t tsid,
 			sepol_security_class_t tclass,
 			sepol_security_id_t * out_sid)
@@ -855,7 +739,7 @@ int sepol_member_sid(sepol_security_id_t ssid,
  * Compute a SID to use for relabeling an object in the 
  * class `tclass' based on a SID pair.  
  */
-int sepol_change_sid(sepol_security_id_t ssid,
+int hidden sepol_change_sid(sepol_security_id_t ssid,
 			sepol_security_id_t tsid,
 			sepol_security_class_t tclass,
 			sepol_security_id_t * out_sid)
@@ -880,11 +764,11 @@ static int validate_perm(hashtab_key_t key, hashtab_datum_t datum, void *p)
 
 	perdatum2 = (perm_datum_t *) hashtab_search(h, key);
 	if (!perdatum2) {
-		printf("security:  permission %s disappeared", key);
+		ERR(NULL, "permission %s disappeared", key);
 		return -1;
 	}
 	if (perdatum->value != perdatum2->value) {
-		printf("security:  the value of permission %s changed", key);
+		ERR(NULL, "the value of permissions %s changed", key);
 		return -1;
 	}
 	return 0;
@@ -906,28 +790,30 @@ static int validate_class(hashtab_key_t key, hashtab_datum_t datum, void *p)
 
 	cladatum2 = (class_datum_t *) hashtab_search(newp->p_classes.table, key);
 	if (!cladatum2) {
-		printf("security:  class %s disappeared\n", key);
+		ERR(NULL, "class %s disappeared", key);
 		return -1;
 	}
 	if (cladatum->value != cladatum2->value) {
-		printf("security:  the value of class %s changed\n", key);
+		ERR(NULL, "the value of class %s changed", key);
 		return -1;
 	}
 	if ((cladatum->comdatum && !cladatum2->comdatum) ||
 	    (!cladatum->comdatum && cladatum2->comdatum)) {
-		printf("security:  the inherits clause for the access vector definition for class %s changed\n", key);
+		ERR(NULL, "the inherits clause for the access "
+			"vector definition for class %s changed", key);
 		return -1;
 	}
 	if (cladatum->comdatum) {
 		if (hashtab_map(cladatum->comdatum->permissions.table, validate_perm,
 				cladatum2->comdatum->permissions.table)) {
-			printf(" in the access vector definition for class %s\n", key);
+			ERR(NULL, " in the access vector definition "
+				"for class %s\n", key);
 			return -1;
 		}
 	}
 	if (hashtab_map(cladatum->permissions.table, validate_perm,
 			cladatum2->permissions.table)) {
-		printf(" in access vector definition for class %s\n", key);
+		ERR(NULL, " in access vector definition for class %s", key);
 		return -1;
 	}
 	return 0;
@@ -952,8 +838,8 @@ static inline int convert_context_handle_invalid_context(
 		sepol_security_context_t s;
 		size_t len;
 
-		context_struct_to_string(context, &s, &len);
-		printf("security:  context %s is invalid\n", s);
+		context_to_string(NULL, policydb, context, &s, &len);
+		ERR(NULL, "context %s is invalid", s);
 		free(s);
 		return 0;
 	}
@@ -1030,9 +916,9 @@ static int convert_context(sepol_security_id_t key __attribute__ ((unused)),
 	return 0;
 
       bad:
-	context_struct_to_string(&oldc, &s, &len);
+	context_to_string(NULL, policydb, &oldc, &s, &len);
 	context_destroy(&oldc);
-	printf("security:  invalidating context %s\n", s);
+	ERR(NULL, "invalidating context %s", s);
 	free(s);
 	return rc;
 }
@@ -1054,14 +940,21 @@ static int convert_context(sepol_security_id_t key __attribute__ ((unused)),
  *
  * Reset the access vector cache.
  */
-int sepol_load_policy(void * data, size_t len)
+int hidden sepol_load_policy(void * data, size_t len)
 {
 	policydb_t oldpolicydb, newpolicydb;
 	sidtab_t oldsidtab, newsidtab;
 	convert_context_args_t args;
 	uint32_t seqno;
 	int rc = 0;
-	struct policy_file file = { 0, data, len }, *fp = &file;
+	struct policy_file file = { 
+	  .type = PF_USE_MEMORY, 
+	  .data = data, 
+	  .len = len,
+	  .fp = NULL }, *fp = &file;
+
+	if (policydb_init(&newpolicydb))
+		return -ENOMEM;
 
 	if (policydb_read(&newpolicydb, fp, 1)) {
 		return -EINVAL;
@@ -1070,8 +963,9 @@ int sepol_load_policy(void * data, size_t len)
 	sepol_sidtab_init(&newsidtab);
 
 	/* Verify that the existing classes did not change. */
-	if (hashtab_map(policydb->p_classes.table, validate_class, &newpolicydb)) {
-		printf("security:  the definition of an existing class changed\n");
+	if (hashtab_map(
+		policydb->p_classes.table, validate_class, &newpolicydb)) {
+		ERR(NULL, "the definition of an existing class changed");
 		rc = -EINVAL;
 		goto err;
 	}
@@ -1118,7 +1012,7 @@ err:
  * the file system and the `file_sid' SID is returned
  * for all files within that file system.
  */
-int sepol_fs_sid(char *name,
+int hidden sepol_fs_sid(char *name,
 		    sepol_security_id_t * fs_sid,
 		    sepol_security_id_t * file_sid)
 {
@@ -1161,7 +1055,7 @@ int sepol_fs_sid(char *name,
  * Return the SID of the port specified by
  * `domain', `type', `protocol', and `port'.
  */
-int sepol_port_sid(uint16_t domain __attribute__ ((unused)),
+int hidden sepol_port_sid(uint16_t domain __attribute__ ((unused)),
 		      uint16_t type __attribute__ ((unused)),
 		      uint8_t protocol,
 		      uint16_t port,
@@ -1204,7 +1098,7 @@ int sepol_port_sid(uint16_t domain __attribute__ ((unused)),
  * the default SID for messages received on the
  * interface.
  */
-int sepol_netif_sid(char *name,
+int hidden sepol_netif_sid(char *name,
 		       sepol_security_id_t * if_sid,
 		       sepol_security_id_t * msg_sid)
 {
@@ -1261,7 +1155,7 @@ static int match_ipv6_addrmask(uint32_t *input, uint32_t *addr, uint32_t *mask)
  * in bytes and `domain' is the communications domain or
  * address family in which the address should be interpreted.
  */
-int sepol_node_sid(uint16_t domain,
+int hidden sepol_node_sid(uint16_t domain,
 		      void *addrp,
 		      size_t addrlen,
 		      sepol_security_id_t *out_sid)
@@ -1335,7 +1229,7 @@ int sepol_node_sid(uint16_t domain,
  */
 #define SIDS_NEL 25
 
-int sepol_get_user_sids(sepol_security_id_t fromsid,
+int hidden sepol_get_user_sids(sepol_security_id_t fromsid,
 	                   char *username,
 			   sepol_security_id_t **sids,
 			   uint32_t *nel)
@@ -1348,6 +1242,7 @@ int sepol_get_user_sids(sepol_security_id_t fromsid,
 	struct sepol_av_decision avd;
 	int rc = 0;
 	unsigned int i, j, reason;
+	ebitmap_node_t *rnode, *tnode;
 
 	fromcon = sepol_sidtab_search(sidtab, fromsid);
 	if (!fromcon) {
@@ -1370,19 +1265,19 @@ int sepol_get_user_sids(sepol_security_id_t fromsid,
 	}
 	memset(mysids, 0, maxnel*sizeof(sepol_security_id_t));
 
-	for (i = ebitmap_startbit(&user->roles); i < ebitmap_length(&user->roles); i++) {
-		if (!ebitmap_get_bit(&user->roles, i)) 
+	ebitmap_for_each_bit(&user->roles.roles, rnode, i) {
+		if (!ebitmap_node_get_bit(rnode, i)) 
 			continue;		
 		role = policydb->role_val_to_struct[i];
 		usercon.role = i+1;
-		for (j = ebitmap_startbit(&role->types); j < ebitmap_length(&role->types); j++) {
-			if (!ebitmap_get_bit(&role->types, j)) 
+		ebitmap_for_each_bit(&role->types.types, tnode, j) {
+			if (!ebitmap_node_get_bit(tnode, j)) 
 				continue;	
 			usercon.type = j+1;
 			if (usercon.type == fromcon->type)
 				continue;
 
-			if (mls_setup_user_range(fromcon, user, &usercon))
+			if (mls_setup_user_range(fromcon, user, &usercon, policydb->mls))
 				continue;
 
 			rc = context_struct_compute_av(fromcon, &usercon, 
@@ -1428,7 +1323,7 @@ out:
  * that cannot support a persistent label mapping or use another
  * fixed labeling behavior like transition SIDs or task SIDs.
  */
-int sepol_genfs_sid(const char *fstype,
+int hidden sepol_genfs_sid(const char *fstype,
 	               char *path,
 		       sepol_security_class_t sclass,
 		       sepol_security_id_t *sid)
@@ -1476,7 +1371,7 @@ out:
 	return rc;
 }
 
-int sepol_fs_use(
+int hidden sepol_fs_use(
 	const char *fstype,
 	unsigned int *behavior,
 	sepol_security_id_t *sid)
